/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.client.guide.parts.contents;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.time.LocalDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;

import net.minecraft.client.gui.FontRenderer;
import net.minecraft.client.gui.GuiTextField;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.profiler.Profiler;
import net.minecraft.util.text.TextFormatting;

import net.minecraftforge.fml.common.registry.ForgeRegistries;

import buildcraft.api.BCModules;
import buildcraft.api.core.BCLog;

import buildcraft.lib.BCLib;
import buildcraft.lib.client.guide.GuiGuide;
import buildcraft.lib.client.guide.GuideManager;
import buildcraft.lib.client.guide.TypeOrder;
import buildcraft.lib.client.guide.font.IFontRenderer;
import buildcraft.lib.client.guide.loader.XmlPageLoader;
import buildcraft.lib.client.guide.parts.GuideChapter;
import buildcraft.lib.client.guide.parts.GuidePageBase;
import buildcraft.lib.client.render.font.ConfigurableFontRenderer;
import buildcraft.lib.gui.GuiIcon;
import buildcraft.lib.gui.pos.GuiRectangle;
import buildcraft.lib.misc.LocaleUtil;
import buildcraft.lib.misc.ProfilerUtil;
import buildcraft.lib.misc.ProfilerUtil.ILogAcceptor;
import buildcraft.lib.misc.RenderUtil;

import buildcraft.lib.client.guide.parts.GuidePart.PagePosition;

/** The base menu for showing all the locations. Should never be registered with and guide managers, this is special and
 * controls them all. */
public class GuidePageContents extends GuidePageBase {
    private static final int ORDER_OFFSET_X = -50;
    private static final int ORDER_OFFSET_Y = 14;

    private ContentsNodeGui contents;
    private final GuiTextField searchText;
    private String lastSearchText = "";

    public GuidePageContents(GuiGuide gui) {
        super(gui);
        loadMainGui();
        FontRenderer fr = new ConfigurableFontRenderer(gui.field_146297_k.field_71466_p).disableShadow();
        searchText = new GuiTextField(0, fr, 0, 0, 80, fr.field_78288_b + 5);
        searchText.func_146185_a(false);
        searchText.func_146193_g(0xFF_00_00_00);
    }

    @Override
    public GuidePageBase createReloaded() {
        GuidePageContents newPage = new GuidePageContents(gui);
        newPage.searchText.func_146180_a(searchText.func_146179_b());
        newPage.searchText.func_146190_e(searchText.func_146198_h());
        newPage.searchText.func_146195_b(searchText.func_146206_l());
        newPage.searchText.func_146199_i(searchText.func_146186_n());
        newPage.numPages = numPages;
        newPage.goToPage(getIndex());
        return newPage;
    }

    public void loadMainGui() {
        contents = GuideManager.INSTANCE.getGuiContents(gui, this, gui.sortingOrder);
    }

    @Override
    public void setFontRenderer(IFontRenderer fontRenderer) {
        super.setFontRenderer(fontRenderer);
        contents.setFontRenderer(fontRenderer);
    }

    @Override
    public List<GuideChapter> getChapters() {
        return contents.getChapters();
    }

    @Override
    public String getTitle() {
        return null;
    }

    @Override
    public void updateScreen() {
        super.updateScreen();
        searchText.func_146178_a();
        if (lastSearchText.equals(searchText.func_146179_b())) {
            if (numPages > 3 && getPage() > numPages) {
                goToPage(numPages);
            } else if (getPage() < 2) {
                searchText.func_146195_b(false);
            }
        } else {
            int __setSize = 0;
            int __listSize = 0;
            Profiler p = new Profiler();
            long start = System.nanoTime();
            p.field_76327_a = true;
            p.func_76320_a("root");
            lastSearchText = searchText.func_146179_b();
            numPages = -1;
            if (lastSearchText.isEmpty()) {
                p.func_76320_a("reset");
                contents.node.resetVisibility();
                p.func_76318_c("invalidate");
                contents.invalidate();
                p.func_76319_b();
            } else {
                p.func_76320_a("search");
                String text = lastSearchText.toLowerCase(Locale.ROOT);
                List<PageLink> ret = GuideManager.INSTANCE.quickSearcher.func_194055_a(text);
                __listSize = ret.size();
                p.func_76318_c("make_set");
                Set<PageLink> matches = new HashSet<>(ret);
                __setSize = matches.size();
                p.func_76318_c("setVisible");
                contents.node.setVisible(matches);
                p.func_76318_c("invalidate");
                contents.invalidate();

                p.func_76318_c("check_isVisible");
                if (contents.node.isVisible()) {
                    searchText.func_146193_g(0xFF_00_00_00);
                } else {
                    searchText.func_146193_g(0xFF_FF_00_00);
                }
                p.func_76319_b();
            }
            p.func_76320_a("refresh_chapters");
            gui.refreshChapters();
            p.func_76319_b();
            p.func_76319_b();
            long end = System.nanoTime();
            try (BufferedWriter writer = Files.newBufferedWriter(new File("bc_perf_guide_search.txt").toPath(),
                StandardOpenOption.WRITE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.CREATE)) {
                ILogAcceptor<IOException> dest = s -> {
                    writer.write(s);
                    writer.newLine();
                };
                dest.write("INFO:");
                dest.write("Date = " + LocalDateTime.now());
                dest.write("Item Count = " + ForgeRegistries.ITEMS.getValues().size());
                dest.write("Block Count = " + ForgeRegistries.BLOCKS.getValues().size());
                dest.write("Search text = '" + lastSearchText + "'");
                dest.write("Time = " + (end - start) + "ns (" + ((end - start) / 10_000 / 100.0) + "ms)");
                dest.write("List size = " + __listSize);
                dest.write("Set size = " + __setSize);
                dest.write("RESULTS:");
                ProfilerUtil.writeProfilerResults(p, "root", dest);
                dest.write("@end");
            } catch (IOException e) {
                BCLog.logger.warn("[lib.guide] Failed to save guide searching profiler results!", e);
            }
        }
    }

    @Override
    protected void renderPage(int x, int y, int width, int height, int index) {
        IFontRenderer f = getFontRenderer();
        if (index == 0) {
            String text = "BuildCraft";
            float scale = 3;
            int fWidth = (int) (f.getStringWidth(text) * scale);
            GlStateManager.func_179094_E();
            GlStateManager.func_179152_a(scale, scale, 1);
            f.drawString(text, (int) ((x + (width - fWidth) / 2) / scale), (int) ((y + height / 2 - 62) / scale), 0);
            GlStateManager.func_179121_F();

            text = "v" + BCLib.VERSION;
            fWidth = f.getStringWidth(text);
            f.drawString(text, x + (width - fWidth) / 2, y + height / 2 - 36, 0);

            scale = 1.5f;
            text = LocaleUtil.localize("options.title");
            fWidth = (int) (f.getStringWidth(text) * scale);
            GlStateManager.func_179094_E();
            GlStateManager.func_179152_a(scale, scale, 1);
            f.drawString(text, (int) ((x + (width - fWidth) / 2) / scale), (int) ((y + height / 2 - 4) / scale), 0);
            GlStateManager.func_179121_F();

            text = XmlPageLoader.SHOW_LORE ? "Show Lore [x]" : "Show Lore [ ]";
            fWidth = f.getStringWidth(text);
            f.drawString(text, x + (width - fWidth) / 2, y + height / 2 + 12, 0);

            text = XmlPageLoader.SHOW_HINTS ? "Show Hints [x]" : "Show Hints [ ]";
            fWidth = f.getStringWidth(text);
            f.drawString(text, x + (width - fWidth) / 2, y + height / 2 + 26, 0);
        } else if (index == 1) {
            int _height = GuideManager.loadedMods.size() + 1;
            if (GuideManager.loadedOther.size() > 0) {
                _height++;
                _height += GuideManager.loadedOther.size();
            }
            int perLineHeight = f.getFontHeight("Ly") + 3;
            _height *= perLineHeight;
            int _y = y + (height - _height) / 2;

            drawCenteredText(TextFormatting.BOLD + "Loaded Mods:", x, _y, width);
            _y += perLineHeight;
            boolean first = true;
            for (String text : GuideManager.loadedMods) {

                if (first && text.contains("modules")) {
                    // BuildCraft
                    double mouseX = gui.mouse.getX();
                    double mouseY = gui.mouse.getY();
                    if (mouseX >= x + width / 2 && mouseX <= x + width * 7 / 8//
                        && mouseY >= _y && mouseY <= _y + perLineHeight) {
                        gui.tooltip.add(LocaleUtil.localize("buildcraft.guide.contents.loaded_modules"));
                        for (BCModules module : BCModules.getLoadedModules()) {
                            gui.tooltip.add((GuideManager.buildCraftModules.contains(module) ? " + " : " - ")
                                + StringUtils.capitalize(module.lowerCaseName));
                        }
                        if (BCModules.getMissingModules().length > 0) {
                            gui.tooltip.add(LocaleUtil.localize("buildcraft.guide.contents.missing_modules"));
                            for (BCModules module : BCModules.getMissingModules()) {
                                gui.tooltip.add(" - " + StringUtils.capitalize(module.lowerCaseName));
                            }
                        }
                    }
                }

                drawCenteredText(text, x, _y, width);
                _y += perLineHeight;
                first = false;
            }
            if (GuideManager.loadedOther.size() > 0) {
                drawCenteredText(TextFormatting.BOLD + "Loaded Resource Packs:", x, _y, width);
                _y += perLineHeight;
                for (String text : GuideManager.loadedOther) {
                    drawCenteredText(text, x, _y, width);
                    _y += perLineHeight;
                }
            }
        } else if (index % 2 == 0) {
            searchText.field_146209_f = x + 23;
            searchText.field_146210_g = y - 23;
            if (!searchText.func_146206_l() && searchText.func_146179_b().isEmpty()) {
                GuiGuide.SEARCH_TAB_CLOSED.drawAt(x + 8, y - 20);
                GuiGuide.SEARCH_ICON.drawAt(x + 8, y - 19);
            } else {
                GuiGuide.SEARCH_TAB_OPEN.drawAt(x - 2, y - 29);
                GuiGuide.SEARCH_ICON.drawAt(x + 8, y - 25);
            }
            searchText.func_146194_f();
        }
        RenderUtil.setGLColorFromInt(-1);
        PagePosition pos = new PagePosition(2, 0);

        pos = contents.render(x, y, width, height, pos, index);

        if (numPages == -1) {
            numPages = pos.page + 1;
        }
        super.renderPage(x, y, width, height, index);
        if (index % 2 == 0) {
            int oX = x + ORDER_OFFSET_X;
            int oY = y + ORDER_OFFSET_Y;
            int i = 0;
            for (int j = 0; j < GuiGuide.ORDERS.length; j++) {
                GuiIcon icon = GuiGuide.ORDERS[j];
                if (gui.sortingOrder == GuiGuide.SORTING_TYPES[j]) {
                    icon = icon.offset(0, 14);
                }
                icon.drawAt(oX, oY);
                oY += 14;
                i++;
            }
        }
    }

    private void drawCenteredText(String text, int x, int y, int width) {
        IFontRenderer f = getFontRenderer();
        int fWidth = f.getStringWidth(text);
        f.drawString(text, (x + (width - fWidth) / 2), y, 0);
    }

    private void drawScaledCenteredText(float scale, String text, int x, int y, int width) {
        IFontRenderer f = getFontRenderer();
        int fWidth = (int) (f.getStringWidth(text) * scale);
        if (scale != 1) {
            GlStateManager.func_179094_E();
            GlStateManager.func_179152_a(scale, scale, 1);
        }
        f.drawString(text, (int) ((x + (width - fWidth) / 2) / scale), (int) (y / scale), 0);
        if (scale != 1) {
            GlStateManager.func_179121_F();
        }
    }

    @Override
    public void handleMouseClick(int x, int y, int width, int height, int mouseX, int mouseY, int mouseButton,
        int index, boolean isEditing) {
        super.handleMouseClick(x, y, width, height, mouseX, mouseY, mouseButton, index, isEditing);
        if (index % 2 == 0) {
            int oX = x + ORDER_OFFSET_X;
            int oY = y + ORDER_OFFSET_Y;
            for (TypeOrder order : GuiGuide.SORTING_TYPES) {
                GuiRectangle rect = new GuiRectangle(oX, oY, 14, 14);
                if (rect.contains(gui.mouse)) {
                    gui.sortingOrder = order;
                    loadMainGui();
                    gui.refreshChapters();
                    contents.setFontRenderer(getFontRenderer());
                    return;
                }
                oY += 14;
            }
            if (!searchText.func_146192_a(mouseX, mouseY, mouseButton) && !searchText.func_146206_l()
                && new GuiRectangle(x - 2, y - 34, 40, 34).contains(mouseX, mouseY)) {
                searchText.func_146195_b(true);
            }
            if (mouseButton == 1 && mouseX >= searchText.field_146209_f && mouseX < searchText.field_146209_f + searchText.field_146218_h
                && mouseY >= searchText.field_146210_g && mouseY < searchText.field_146210_g + searchText.field_146219_i) {
                searchText.func_146180_a("");
            }
        }
        if (mouseButton == 0) {
            if (index == 0) {
                IFontRenderer f = getFontRenderer();
                String text = XmlPageLoader.SHOW_LORE ? "Show Lore [x]" : "Show Lore [ ]";
                int fWidth = f.getStringWidth(text);
                GuiRectangle rect;
                rect = new GuiRectangle(x + (width - fWidth) / 2, y + height / 2 + 12, fWidth, f.getFontHeight(text));
                if (rect.contains(mouseX, mouseY)) {
                    XmlPageLoader.SHOW_LORE = !XmlPageLoader.SHOW_LORE;
                }

                text = XmlPageLoader.SHOW_HINTS ? "Show Hints [x]" : "Show Hints [ ]";
                fWidth = f.getStringWidth(text);
                rect = new GuiRectangle(x + (width - fWidth) / 2, y + height / 2 + 26, fWidth, f.getFontHeight(text));
                if (rect.contains(mouseX, mouseY)) {
                    XmlPageLoader.SHOW_HINTS = !XmlPageLoader.SHOW_HINTS;
                }
            }
        }

        if (new GuiRectangle(x, y, width, height).contains(mouseX, mouseY)) {
            contents.onClicked(x, y, width, height, new PagePosition(2, 0), index);
        }
    }

    @Override
    public boolean keyTyped(char typedChar, int keyCode) throws IOException {
        return searchText.func_146201_a(typedChar, keyCode);
    }
}
